importScripts("./ComPDFkit.js")

import { parseAdobePDFTimestamp, convertToPDFTimestamp, roundToDecimalPlaces, convertColorToCppFormat, convertCppRGBToHex } from '../ui_utils';
import { AnnotationType, WidgetType, LineTypeString, StampType, StampTypeString, TextStampShapeString, ActionTypeString, WidgetTypeString, AnnotationFlags, BorderStyleInt, BorderStyleString, ALIGN, ALIGNMAP } from '../../constants'

let ComPDFKitJS = {}
let DataArray = []
let LineType = {}
let BorderStyle = {}
let PageSize = {}
let Rect = {}
let RGBColor = {}
let FontDa = {}
let PDFDestination = {}
let WidgetItem = {}
let WidgetItemsArray = {}
let password = ''
let TextFindItem = {}
let EditCharPlace = {}
let EndEditCharPlace = {}
let RectArray = []
let CursorPoints = {}
let EditTextStyle = {}
let U8StringData = ''
let PDFRange = {}
let TextRectArray = []

import MessageHandler from "../message_handler"
import { convertFileToBuffer } from '../fileHandler';

async function sleep(delay) {
  await new Promise((resolve) => setTimeout(resolve, delay))
}

function hexToRGB(hexColor) {
  const color = parseInt(hexColor.slice(1), 16);

  const red = (color >> 16) & 0xFF;
  const green = (color >> 8) & 0xFF;
  const blue = color & 0xFF;
  return {
    R: roundToDecimalPlaces(red / 255),
    G: roundToDecimalPlaces(green / 255),
    B: roundToDecimalPlaces(blue / 255)
  }
}

class CPDFWorker {
  static async setup(doc) {
    this.doc = doc
    const messageHandler = new MessageHandler('worker', 'main', self)
    this.MessageHandler = messageHandler

    let testMessageProcessed = false

    messageHandler.on("test", function (data) {
      if (testMessageProcessed) {
        return; // we already processed 'test' message once
      }
      testMessageProcessed = true;

      // Ensure that `TypedArray`s can be sent to the worker.
      messageHandler.send("test", data instanceof Uint8Array);
    });

    messageHandler.send("ready", null)

    messageHandler.send("initDoc", { doc })

    messageHandler.on("Ready", function (data) {
      return CPDFWorker.createDocumentHandler(self)
    })
  }

  static createDocumentHandler(port) {
    const messageHandler = this.MessageHandler

    messageHandler.on('LoadFile', (data) => {
      ComPDFKitJS.opened_files = []
      ComPDFKitJS.opened_files[0] = data.buffer
      return true
    })

    messageHandler.on('LoadDocumentByStream', (data) => {
      const { doc, fileId, length, password: rawPassword } = data
      const password = stringToNewUTF8(rawPassword)
      return Module._LoadDocumentByStream(doc, fileId, length, password)
    })

    messageHandler.on('GetPageCount', (data) => {
      const { doc } = data
      return Module._GetPageCount(doc)
    })

    messageHandler.on('GetPageSize', (data) => {
      const { doc, pageIndex } = data

      PageSize.Width = 0
      PageSize.Height = 0

      Module._GetPageSizeForIndex(doc, pageIndex)
      const pagePtr = Module._GetPage(doc, pageIndex)

      return {
        pagePtr,
        width: PageSize.Width,
        height: PageSize.Height
      }
    })

    messageHandler.on('InitAnnot', (data) => {
      const { pagePtr } = data
      const annotCount = Module._GetAnnotCount(pagePtr)
      const annotations = []
      for (let i = 0; i < annotCount; i++) {
        const annotPtr = Module._InitAnnot(pagePtr, i)
        const type = Module._GetAnnotType(annotPtr)
        const annotation = {
          annotPtr,
          type,
          pagePtr
        }
        annotations.push(annotation)
      }

      return {
        annotCount,
        annotations
      }
    })

    messageHandler.on('GetAnnotation', (data) => {
      const { doc, pagePtr, annotPtr, typeInt, scale } = data

      return getAnnotation({
        doc,
        pagePtr,
        annotPtr,
        typeInt,
        scale
      })
    })

    messageHandler.on('GetRenderAnnot', (data) => {
      const { annotPtr, annotation, scale } = data
      const width = parseInt(annotation.rect.right * scale - annotation.rect.left * scale + 1)
      const height = parseInt(annotation.rect.bottom * scale - annotation.rect.top * scale + 1)
      let pixelNum = width * height

      let imageBytes = pixelNum * 4
      let imageptr = _malloc(imageBytes)
      for (var i = 0; i < pixelNum; i++) {
        Module.HEAP32[imageptr / 4 + i] = 0
      }

      Module._RenderAnnot(annotPtr, width, height, imageptr, 0)
      let imageArray = new Uint8Array(imageBytes)
      for (var i = 0; i < imageBytes; i += 4) {
        //bgra 转 rgba
        imageArray[i] = Module.HEAPU8[imageptr + i + 2]
        imageArray[i + 1] = Module.HEAPU8[imageptr + i + 1]
        imageArray[i + 2] = Module.HEAPU8[imageptr + i]
        imageArray[i + 3] = Module.HEAPU8[imageptr + i + 3]
      }
      _free(imageptr)
      return imageArray
    })

    messageHandler.on('GetWidgetAnnotation', (data) => {
      const { doc, pagePtr, annotPtr, typeInt } = data

      return getWidgetAnnotation({
        doc,
        pagePtr,
        annotPtr,
        typeInt
      })
    })

    messageHandler.on('PushRenderTask', (data) => {
      const { pagePtr, scale, left, right, bottom, top } = data

      let param = {
        pagePtr,
        scale,
        left,
        right,
        bottom,
        top,
        mode: -1,
        flag: 0,
        form: 0
      }
      let imageArray = RenderPageBitmapWithMatrix(param)
      return { imageArray }
    })

    messageHandler.on('EditAnnotation', (data) => {
      const { annotation } = data
      if (annotation.rect) {
        setAnnotRect(annotation)
      }
      if (annotation.color) {
        setAnnotRGBColor(annotation)
      }
      if (annotation.transparency) {
        setAnnotTransparency(annotation)
      }
      if (annotation.content && annotation.operate === 'mod-form') {
        setTextFieldContent(annotation)
      }
      if (annotation.content && annotation.operate === 'mod-annot') {
        setAnnotContent(annotation)
      }
      if (annotation.linePoints) {
        setLinePoint(annotation)
      }
      if (annotation.inkPointes) {
        setInkPointes(annotation)
      }
      if (annotation.destPage && annotation.operate === 'mod-annot') {
        setLinkDest(annotation)
      }
      if (annotation.destPage && annotation.operate === 'mod-form') {
        setPushButtonDest(annotation)
      }
      if (annotation.url && annotation.operate === 'mod-annot') {
        setLinkUri(annotation)
      }
      if (annotation.url && annotation.operate === 'mod-form') {
        setPushButtonUri(annotation)
      }
      if (annotation.isHidden || annotation.isHidden === 0) {
        setAnnotHidden(annotation)
      }
      if (annotation.isChecked || annotation.isChecked === 0) {
        if (annotation.type === 'checkbox') {
          setCheckBoxIsChecked(annotation)
        } else if (annotation.type === 'radiobutton') {
          setRadioButtonIsChecked(annotation)
        }
      }
      if (annotation.checkStyle) {
        setCheckStyle(annotation)
      }
      if (annotation.backgroundColor && annotation.operate === 'mod-form') {
        setWidgetBackgroundColor(annotation)
      }
      if (annotation.title && annotation.operate === 'mod-form') {
        const { annotPtr, title: rawTitle } = annotation
        const title = stringToNewUTF8(rawTitle)
        Module._SetButtonTitle(annotPtr, title)
      }
      if (annotation.fieldName) {
        setWidgetFieldName(annotation)
      }
      if (annotation.fontName || annotation.fontSize || annotation.color) {
        setAnnotFontData(annotation)
      }
      if (annotation.selected || annotation.selected === 0) {
        setAnnotSelected(annotation)
      }
      if (annotation.item) {
        addWidgetItem(annotation)
      }
      if (annotation.moveOption) {
        setWidgetItems(annotation)
      }
      if (annotation.deleteItemIndex) {
        deleteWidgetItem(annotation)
      }
      if (annotation.textAlignment) {
        const { annotPtr, textAlignment } = annotation
        const textAlignmentInt = ALIGNMAP[textAlignment.toUpperCase()]
        Module._SetWidgetTextJustification(annotPtr, textAlignmentInt)
      }
      if (annotation.isMultiLine) {
        const { annotPtr, isMultiLine } = annotation
        Module._SetMultiLine(annotPtr, isMultiLine)
      }
    })

    messageHandler.on('FlattenPage', (data) => {
      const { doc, saveType, password, oldPassword } = data
      return flattenPdfDownload({
        doc,
        saveType,
        password,
        oldPassword
      })
    })

    messageHandler.on('CreateAnnotation', (data) => {
      const { doc, pagePtr, annotation } = data
      if (annotation.operate === 'add-form') {
        return createWidget(doc, pagePtr, annotation)
      }
      return createAnnotation(doc, pagePtr, annotation)
    })

    messageHandler.on('GetAnnotRect', (data) => {
      const { pagePtr, annotPtr } = data

      Rect = {}

      Module._GetAnnotRect(pagePtr, annotPtr)

      return {
        left: Rect.Left,
        top: Rect.Top,
        right: Rect.Right,
        bottom: Rect.Bottom
      }
    })

    messageHandler.on('SetAnnotRect', (data) => {
      const { pagePtr, annotPtr, rect } = data

      Module._SetAnnotRect(pagePtr, annotPtr, rect.left, rect.right, rect.bottom, rect.top)
    })

    messageHandler.on('RemoveAnnot', (data) => {
      const { annotPtr } = data
      const res = Module._RemoveAnnot(annotPtr)
    })

    messageHandler.on('GetCharIndexAtPos', (data) => {
      const { pagePtr, textPtr, point } = data
      const { x, y } = point
      const index = Module._GetCharIndexAtPos(pagePtr, textPtr, x, y, 5, 5)
      return index !== -1
    })

    messageHandler.on('GetCharsRangeAtPos', (data) => {
      const { pagePtr, textPtr, start, end } = data
      PDFRange = {}
      Module._GetCharsRangeAtPos(pagePtr, textPtr, start.x, start.y, end.x, end.y, 5, 5)
      const rawTextContent = Module._GetTextContent(textPtr, PDFRange.Location, PDFRange.Length)
      const textContent = UTF8ToString(rawTextContent)
      TextRectArray = []
      Module._GetRects(pagePtr, textPtr, PDFRange.Location, PDFRange.Length)
      return {
        textContent,
        textRects: TextRectArray
      }
    })

    messageHandler.on('XFDFExportAnnotations', (data) => {
      const { doc } = data
      const xfdfStream = exportXFDFByStream(doc)
      return xfdfStream.Content
    })

    messageHandler.on('XFDFImportAnnotations', (data) => {
      const { doc, xfdfBuffer } = data
      let xfdfBufferPtr = _malloc(xfdfBuffer.length)
      Module.HEAPU8.set(xfdfBuffer, xfdfBufferPtr)
      const res = Module._XFDFImportAnnotaitons(doc, xfdfBufferPtr, xfdfBuffer.length)
      _free(xfdfBufferPtr)
    })

    messageHandler.on('SaveDocumentByStream', (data) => {
      return saveDocumentByStream(data)
    })

    messageHandler.on('SetPassword', (data) => {
      const { doc, password: rawPassword } = data
      const password = stringToNewUTF8(rawPassword)
      const res = Module._SetPassword(doc, password, '', '', '', '', '', '')
    })

    messageHandler.on('CompareDocument', async ({ data }) => {
      const { leftFile, rightFile, leftFilePassword: rawLeftFilePassword, rightFilePassword: rawRightFilePassword } = data
      const oldBuffer = await convertFileToBuffer(leftFile)
      const oldDocPtr = Module._InitDocument()
      ComPDFKitJS.opened_files[1] = oldBuffer
      const leftFilePassword = stringToNewUTF8(rawLeftFilePassword)
      Module._LoadDocumentByStream(oldDocPtr, 1, oldBuffer.length, leftFilePassword)
      const oldPagesCount = Module._GetPageCount(oldDocPtr)

      const newBuffer = await convertFileToBuffer(rightFile)
      const newDocPtr = Module._InitDocument()
      ComPDFKitJS.opened_files[2] = newBuffer
      const rightFilePassword = stringToNewUTF8(rawLeftFilePassword)
      Module._LoadDocumentByStream(newDocPtr, 2, newBuffer.length, rightFilePassword)
      const newPagesCount = Module._GetPageCount(newDocPtr)

      const minPageCount = Math.min(oldPagesCount, newPagesCount)

      // type 1: 内容对比 2: 覆盖对比
      if (data.type === 1) {
        const { replaceColor: rawReplaceColor, insertColor: rawInsertColor, deleteColor: rawDeleteColor, textCompare, imgCompare } = data

        const comparePtr = Module._InitCompare(oldDocPtr, newDocPtr)

        // 对比类型 0: text 1: img 2: text & img
        let compareType = 0
        if (textCompare && imgCompare) {
          compareType = 2
        } else if (imgCompare) {
          compareType = 1
        }

        // 对比的颜色
        const replaceColor = hexToRGB(rawReplaceColor)
        const insertColor = hexToRGB(rawInsertColor)
        const deleteColor = hexToRGB(rawDeleteColor)
        Module._SetReplaceResColor(comparePtr, replaceColor.R * 255, replaceColor.G * 255, replaceColor.B * 255)
        Module._SetInsertResColor(comparePtr, insertColor.R * 255, insertColor.G * 255, insertColor.B * 255)
        Module._SetDeleteResColor(comparePtr, deleteColor.R * 255, deleteColor.G * 255, deleteColor.B * 255)

        // 对比的透明度
        Module._SetReplaceResTransparency(comparePtr, 0.5 * 255)
        Module._SetInsertResTransparency(comparePtr, 0.5 * 255)
        Module._SetDeleteResTransparency(comparePtr, 0.5 * 255)
        for (let i = 0; i < minPageCount; i++) {
          const compareResultsPtr = Module._DoCompare(comparePtr, i, i, compareType, 1)
          // 对比的结果数量
          Module._GetCompareResultsCount(compareResultsPtr, 1)

          // 对比的结果位置
          CompareResultArray = []
          Module._GetCompareResults(oldDocPtr, newDocPtr, compareResultsPtr, 1)

          // 对比的replace、insert、delete的数量
          ResultsCountParams = {}
          Module._GetResultsCountParams(compareResultsPtr)
        }

        // 获取对比结果文档 buffer
        DataArray = []
        Module._GenerateNewDoc(comparePtr)
        let pdfData = []
        for (let i = 0; i < DataArray.length; i++) {
          let subArray = new Uint8Array(DataArray[i])
          pdfData.push(subArray)
        }
        DataArray = []
        const blobData = new Blob(pdfData, { type: "application/pdf" })
        const oldDocument = saveDocument(oldDocPtr, 2)
        const newDocument = saveDocument(newDocPtr, 2)
        return [
          oldDocument,
          newDocument,
          blobData
        ]
      } else if (data.type === 2) {
        const { inColor: rawInColor, newColor: rawNewColor, inTransparency: rawInTransparency, newTransparency: rawNewTransparency, coverType } = data

        const compareDrawingsPtr = Module._InitCompareDrawing(oldDocPtr, newDocPtr)

        // 设置对比的类型
        Module._SetBlendMode(compareDrawingsPtr, coverType)
        Module._SetNoFill(compareDrawingsPtr, 1)
        // 透明度值为0-1
        const inTransparency = rawInTransparency / 100
        const newTransparency = rawNewTransparency / 100
        // 设置对比的透明度
        Module._SetStrokeAlphaCompare(compareDrawingsPtr, inTransparency)
        Module._SetStrokeAlphaCompared(compareDrawingsPtr, newTransparency)
        Module._SetFillAlphaCompare(compareDrawingsPtr, inTransparency)
        Module._SetFillAlphaCompared(compareDrawingsPtr, newTransparency)
        // 设置对比的颜色
        const inColor = hexToRGB(rawInColor)
        const newColor = hexToRGB(rawNewColor)
        Module._SetColorCompare(compareDrawingsPtr, inColor.R * 255, inColor.G * 255, inColor.B * 255)
        Module._SetColorCompared(compareDrawingsPtr, newColor.R * 255, newColor.G * 255, newColor.B * 255)

        Module._CompareContent(compareDrawingsPtr)
        // 获取对比结果文档 buffer
        const docPtr = Module._GetDocGenerate(compareDrawingsPtr)
        const blobData = saveDocumentByStream({
          doc: docPtr,
        })
        return blobData
      }
    })

    messageHandler.on('GetOutlines', (data) => {
      const { doc } = data

      const rootOutlinePtr = Module._GetOutlineRoot(doc)

      const outlines = []
      getOutlines(doc, rootOutlinePtr, outlines, -1)
      return outlines
    })

    messageHandler.on('InitText', (data) => {
      const { pagePtr } = data
      const textPtr = Module._InitText(pagePtr)

      return textPtr
    })

    messageHandler.on('ClearText', (data) => {
      const { textPtr } = data
      Module._ClearText(textPtr)
    })

    messageHandler.on('ClearPage', (data) => {
      const { pagePtr } = data
      Module._ClearPage(pagePtr)
    })

    messageHandler.on('Search', (data) => {
      const { pagesPtr, value } = data
      const text = stringToNewUTF8(value)

      const searchResults = []
      for (let i = 0; i < pagesPtr.length; i++) {
        const ptr = pagesPtr[i]
        const pagePtr = ptr.pagePtr
        const textPtr = ptr.textPtr
        const textFindPtr = Module._InitTextFind(textPtr)
        const res = Module._FindStart(textPtr, textFindPtr, text, 0, 0)

        let resFind = 1
        let pageSearchIndex = 0
        do {
          TextFindItem = {}
          resFind = Module._FindNext(pagePtr, textPtr, textFindPtr)

          if (resFind) {
            const { Left: left, Top: top, Right: right, Bottom: bottom, Content: content } = TextFindItem
            searchResults.push({
              pageNum: i + 1,
              left,
              top,
              right,
              bottom,
              content,
              pageSearchIndex,
              searchValue: value
            })
            pageSearchIndex++
          }
        } while (resFind)
      }

      return searchResults
    })

    messageHandler.on('InitEditPage', (data) => {
      const { pagePtr } = data
      const editPagePtr = Module._InitEditPage(pagePtr)
      return { editPagePtr, pagePtr }
    })

    messageHandler.on('BeginEdit', (data) => {
      const { editPagePtr, type } = data
      Module._BeginEdit(editPagePtr, type)
    })

    messageHandler.on('EndEdit', (editPagePtr) => {
      Module._EndEdit(editPagePtr)
    })

    messageHandler.on('GetEditAreaCount', (editPagePtr) => {
      return Module._GetEditAreaCount(editPagePtr)
    })

    messageHandler.on('GetEditArea', (data) => {
      const { editPagePtr, index } = data
      return Module._GetEditArea(editPagePtr, index)
    })

    messageHandler.on('IsTextArea', (editAreaPtr) => {
      return Module._IsTextArea(editAreaPtr)
    })

    messageHandler.on('SetFrame', (data) => {
      const { pagePtr, editAreaPtr, rect } = data

      Module._SetFrame(
        pagePtr,
        editAreaPtr,
        rect.left,
        rect.right,
        rect.bottom,
        rect.top
      )
    })

    messageHandler.on('GetFrame', (data) => {
      const { pagePtr, editAreaPtr } = data

      Rect = {}

      Module._GetFrame(pagePtr, editAreaPtr)

      return {
        left: Rect.Left,
        top: Rect.Top,
        right: Rect.Right,
        bottom: Rect.Bottom
      }
    })

    messageHandler.on('GetBeginAndEndCharPlace', (editAreaPtr) => {
      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetBeginAndEndCharPlace(editAreaPtr)

      return {
        start: EditCharPlace,
        end: EndEditCharPlace
      }
    })

    messageHandler.on('GetTextByRange', (data) => {
      const { editAreaPtr, start, end } = data
      Module._GetTextByRange(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
      )
      return U8StringData
    })

    messageHandler.on('GetText', async (editAreaPtr) => {

      EditCharPlace = {}
      EndEditCharPlace = {}

      await Module._GetBeginAndEndCharPlace(editAreaPtr)

      await Module._GetTextByRange(
        editAreaPtr,
        EditCharPlace.SectionIndex, EditCharPlace.LineIndex, EditCharPlace.RunIndex, EditCharPlace.CharIndex,
        EndEditCharPlace.SectionIndex, EndEditCharPlace.LineIndex, EndEditCharPlace.RunIndex, EndEditCharPlace.CharIndex,
      )
      return U8StringData
    })

    messageHandler.on('SelectCharsRangeAtPos', (data) => {
      const { pagePtr, editAreaPtr, startPoint, endPoint } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._SelectCharsRangeAtPos(
        pagePtr,
        editAreaPtr,
        startPoint.x, startPoint.y,
        endPoint.x, endPoint.y,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetSectionAtPos', (data) => {
      const { pagePtr, editAreaPtr, pointX, pointY } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetSectionAtPos(
        pagePtr,
        editAreaPtr,
        pointX,
        pointY
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetTextCursorPoints', (data) => {
      const { pagePtr, editAreaPtr, char } = data

      CursorPoints = {}

      Module._GetTextCursorPoints(
        pagePtr,
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return CursorPoints
    })

    messageHandler.on('GetPrevCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetPrevCharPlace(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetNextCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetNextCharPlace(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetUpCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetUpCharPlace(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('GetDownCharPlace', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}
      EndEditCharPlace = {}

      Module._GetDownCharPlace(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return { start: EditCharPlace, end: EndEditCharPlace }
    })

    messageHandler.on('InsertText', (data) => {
      const { editAreaPtr, char, text } = data

      EditCharPlace = {}

      let textPtr = stringToNewUTF8(text)

      Module._InsertText(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
        textPtr,
        false
      )
      return EditCharPlace
    })

    messageHandler.on('BackSpaceChar', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}

      Module._BackSpaceChar(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
        true
      )
      return EditCharPlace
    })

    messageHandler.on('DeleteChar', (data) => {
      const { editAreaPtr, char } = data

      EditCharPlace = {}

      Module._DeleteChar(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
        true
      )
      return EditCharPlace
    })

    messageHandler.on('DeleteChars', (data) => {
      const { editAreaPtr, start, end } = data

      EditCharPlace = {}

      Module._DeleteChars(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
        true
      )
      return EditCharPlace
    })

    messageHandler.on('GetSelectLineRectsCounts', (data) => {
      const { editAreaPtr, start, end } = data
      return Module._GetSelectLineRectsCounts(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
      )
    })

    messageHandler.on('GetSelectLineRects', (data) => {
      const { pagePtr, editAreaPtr, start, end } = data

      RectArray = []

      Module._GetSelectLineRects(
        pagePtr,
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex
      )
      return RectArray
    })

    messageHandler.on('SetCharsFontColor', (data) => {
      const { editAreaPtr, start, end, color } = data
      Module._SetCharsFontColor(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
        color.r, color.g, color.b
      )
    })

    messageHandler.on('SetCharsFontTransparency', (data) => {
      const { editAreaPtr, start, end, opacity } = data
      Module._SetCharsFontTransparency(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
        opacity
      )
    })

    messageHandler.on('SetTextAligningSection', (data) => {
      const { editAreaPtr, char, alignType } = data
      Module._SetTextAligningSection(
        editAreaPtr,
        alignType,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex
      )
    })

    messageHandler.on('SetTextAligningRange', (data) => {
      const { editAreaPtr, start, end, alignType } = data
      Module._SetTextAligningRange(
        editAreaPtr,
        alignType,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex
      )
    })

    messageHandler.on('SetCharsFontSize', (data) => {
      const { editAreaPtr, start, end, fontSize } = data
      Module._SetCharsFontSize(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
        fontSize,
        true
      )
    })

    messageHandler.on('SetFontFromNativeTrueTypeFont', (data) => {
      const { editAreaPtr, start, end, fontFamily } = data

      const font = stringToNewUTF8(fontFamily)
      Module._SetFontFromNativeTrueTypeFont(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex,
        font
      )
    })

    messageHandler.on('RefreshRange', (data) => {
      const { editAreaPtr, start, end } = data
      EditCharPlace = {}
      EndEditCharPlace = {}
      Module._RefreshRange(
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex
      )

      return {
        start: EditCharPlace,
        end: EndEditCharPlace
      }
    })

    messageHandler.on('SetCharsFontStyle', async (data) => {
      const { editAreaPtr, start, end, fontStyle } = data

      let action = '_' + fontStyle

      Module[action](
        editAreaPtr,
        start.SectionIndex, start.LineIndex, start.RunIndex, start.CharIndex,
        end.SectionIndex, end.LineIndex, end.RunIndex, end.CharIndex
      )
    })

    messageHandler.on('CreateNewTextArea', (data) => {
      const { pagePtr, editPagePtr, rect, fontData, alignType } = data

      return Module._CreateNewTextArea(
        pagePtr,
        editPagePtr,
        rect.left,
        rect.right,
        rect.bottom,
        rect.top,
        stringToNewUTF8(fontData.fontName),
        fontData.fontSize,
        fontData.r,
        fontData.g,
        fontData.b,
        fontData.opacity,
        fontData.isBold,
        fontData.italic,
        alignType
      )
    })

    messageHandler.on('RemoveEditAreaByIndex', (data) => {
      const { editPagePtr, editAreaIndex } = data
      Module._RemoveEditAreaByIndex(
        editPagePtr,
        editAreaIndex
      )
    })

    messageHandler.on('RemoveEditArea', (data) => {
      const { editPagePtr, editAreaPtr } = data
      Module._RemoveEditArea(
        editPagePtr,
        editAreaPtr
      )
    })

    messageHandler.on('CopyArea', (data) => {
      const { editPagePtr, editAreaPtr } = data
      return Module._CopyArea(
        editPagePtr,
        editAreaPtr
      )
    })

    messageHandler.on('GetCopyAreaContent', (data) => {
      const { editPagePtr, editAreaInfoPtr } = data
      Module._GetCopyAreaContent(
        editPagePtr,
        editAreaInfoPtr
      )
      return U8StringData
    })

    messageHandler.on('GetTextStyle', (data) => {
      const { editAreaPtr, char } = data

      EditTextStyle = {}

      Module._GetTextStyle(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return EditTextStyle
    })

    messageHandler.on('GetBaseFontName', (data) => {
      const { editAreaPtr, char } = data

      Module._GetBaseFontName(
        editAreaPtr,
        char.SectionIndex, char.LineIndex, char.RunIndex, char.CharIndex,
      )
      return U8StringData
    })

    messageHandler.on('SetFontCallBackForEditPage', (data) => {
      const { editPagePtr, file } = data

      let reader = new FileReader()
      reader.onload = async (e) => {
        let buf = new Uint8Array(e.target.result)
        ComPDFKitJS.opened_Font = []
        ComPDFKitJS.opened_FontNames = []
        let cur_file_id = 0
        cur_file_id = ComPDFKitJS.opened_Font.length
        ComPDFKitJS.opened_Font[cur_file_id] = buf
        ComPDFKitJS.opened_FontNames[cur_file_id] = 'DroidSansFallbackFull'

        Module._SetFontCallBackForEditPage(editPagePtr)
      }
      reader.readAsArrayBuffer(file)
    })
  }

}

async function initDoc() {
  while (!runtimeInitialized) {
    await sleep(100)
  }
  return new Promise((resolve, reject) => {
    const doc = Module._InitDocument()
    resolve(doc)
  })
}

initDoc().then((doc) => {
  CPDFWorker.setup(doc)
})

function setPassword(data) {
  const { doc, password: rawPassword } = data
  const password = stringToNewUTF8(rawPassword)
  const res = Module._SetPassword(doc, password, '', '', '', '', '', '')
}

function RenderPageBitmapWithMatrix(data) {
  const { pagePtr, scale, left, right, bottom, top, mode, flag, form } = data
  let pixelNum = parseInt(right - left) * parseInt(bottom - top)
  let imageBytes = pixelNum * 4
  let imageptr = _malloc(imageBytes)
  for (var i = 0; i < pixelNum; i++) {
    Module.HEAP32[imageptr / 4 + i] = 0
  }

  Module._RenderPageBitmapWithMatrix(pagePtr, scale, left, right, bottom, top, mode, imageptr, flag, form)
  let imageArray = new Uint8Array(imageBytes)
  for (var i = 0; i < imageBytes; i += 4) {
    //bgra 转 rgba
    imageArray[i] = Module.HEAPU8[imageptr + i + 2]
    imageArray[i + 1] = Module.HEAPU8[imageptr + i + 1]
    imageArray[i + 2] = Module.HEAPU8[imageptr + i]
    imageArray[i + 3] = Module.HEAPU8[imageptr + i + 3]
  }
  _free(imageptr)

  return imageArray
}

function createAnnotation(doc, pagePtr, annotation) {
  const typeInt = AnnotationType[annotation.type.toUpperCase()]

  const annotPtr = Module._CreateAnnot(doc, pagePtr, typeInt)

  const { color, lineWidth, content, createDate, transparency, fillTransparency, rect } = annotation
  if (color) {
    setAnnotRGBColor({
      annotPtr,
      color
    })
  }

  if (content) {
    setAnnotContent({
      annotPtr,
      content
    })
  }

  if (createDate) {
    setAnnotCreationDate({
      annotPtr,
      createDate
    })
  }

  if (transparency) {
    setAnnotTransparency({
      annotPtr,
      transparency
    })
  }

  if (fillTransparency) {
    SetAnnotFilledTransparency({
      annotPtr,
      fillTransparency
    })
  }

  if (rect) {
    setAnnotRect({
      pagePtr,
      annotPtr,
      rect
    })
  }
  switch (typeInt) {
    case AnnotationType.TEXT:
      break
    case AnnotationType.LINK:
      break
    case AnnotationType.FREETEXT:
      const { bgColor, fontName, fontSize, textAlignment } = annotation
      if (bgColor) {
        Module._SetFreeTextBgColor(annotPtr, bgColor.R / 255, bgColor.G / 255, bgColor.B / 255)
      } else {
        Module._ClearFreeTextBgColor(annotPtr)
      }

      setFreeText({
        annotPtr,
        fontName,
        fontSize,
        color,
        textAlignment,
        content
      })
      break
    case AnnotationType.LINE:
      const { tail, linePoints } = annotation
      if (tail) {
        Module._SetLineType(annotPtr, 0, 1)
      }
      Module._SetLinePoint(pagePtr, annotPtr, linePoints[0], linePoints[1], linePoints[2], linePoints[3])
    case AnnotationType.SQUARE:
    case AnnotationType.CIRCLE:
      createShape({
        annotPtr,
        color,
        lineWidth,
        transparency,
        fillTransparency
      })
      break
    case AnnotationType.HIGHLIGHT:
    case AnnotationType.UNDERLINE:
    case AnnotationType.SQUIGGLY:
    case AnnotationType.STRIKEOUT:
      const { quadPoints } = annotation
      createMarkup({
        pagePtr,
        annotPtr,
        color,
        transparency,
        quadPoints
      })
      break
    case AnnotationType.STAMP:
      const { stampType } = annotation
      if (stampType === 'standard') {
        createStandardStamp({
          annotPtr,
          content
        })
      } else if (stampType === 'text') {
        annotation.annotPtr = annotPtr
        createTextStamp(annotation)
      } else if (stampType === 'image') {
        annotation.annotPtr = annotPtr
        annotation.doc = doc
        createImageStamp(annotation)
      }
      break
    case AnnotationType.INK:
      const { inkPointes } = annotation
      setAnnotBorderWidth({
        annotPtr,
        lineWidth
      })
      setInkPointes({
        pagePtr,
        annotPtr,
        inkPointes
      })
  }

  if (rect) {
    setAnnotRect({
      pagePtr,
      annotPtr,
      rect
    })
  }

  Module._UpdateAnnotAp(annotPtr, 0)
  return annotPtr
}

function createWidget(doc, pagePtr, annotation) {
  const { type, createDate, fieldName, backgroundColor, checkStyle, rect } = annotation

  const typeInt = WidgetType[type.toUpperCase()]
  const annotPtr = Module._CreateWidget(doc, pagePtr, typeInt)

  if (createDate) {
    setAnnotCreationDate({
      annotPtr,
      createDate
    })
  }

  if (backgroundColor) {
    setWidgetBackgroundColor({
      annotPtr,
      backgroundColor
    })
  }

  if (checkStyle) {
    setCheckStyle({
      annotPtr,
      checkStyle
    })
  }

  if (fieldName) {
    setWidgetFieldName({
      annotPtr,
      fieldName
    })
  }

  const { borderColor, borderWidth, borderStyle, isChecked, color, fontName, fontSize } = annotation
  setWidgetBorder({
    annotPtr,
    borderColor,
    borderWidth,
    borderStyle
  })

  switch (type) {
    case 'checkbox':
      if (isChecked || isChecked === 0) {
        setCheckBoxIsChecked({
          annotPtr,
          isChecked
        })
      }
      break
    case 'radiobutton':
      if (isChecked || isChecked === 0) {
        setRadioButtonIsChecked({
          annotPtr,
          isChecked
        })
      }
      break
    case 'listbox':
    case 'combobox':
      const { items, selected } = annotation
      for (let i = 0; i < items.length; i++) {
        addWidgetItem({
          annotPtr,
          item: {
            value: items[i].Value,
            index: i
          }
        })
      }
      setAnnotFontData({
        annotPtr,
        fontName,
        fontSize,
        color
      })
      if (selected || selected === 0) {
        setAnnotSelected({
          annotPtr,
          selected
        })
      }
      break
    case 'pushbutton':
      setAnnotFontData({
        annotPtr,
        fontName,
        fontSize,
        color
      })
      const { title: rawTitle } = annotation
      const title = stringToNewUTF8(rawTitle)
      Module._SetButtonTitle(annotPtr, title)
      break
    case 'textfield':
      const { isMultiLine, textAlignment } = annotation
      const textAlignmentInt = ALIGNMAP[textAlignment.toUpperCase()]
      Module._SetMultiLine(annotPtr, isMultiLine)
      Module._SetWidgetTextJustification(annotPtr, textAlignmentInt)
      Module._SetTextFieldSpe(annotPtr, 0)
      setAnnotFontData({
        annotPtr,
        fontName,
        fontSize,
        color
      })
      break
  }

  setAnnotRect({
    pagePtr,
    annotPtr,
    rect
  })

  Module._UpdateFormAp(annotPtr)
  return annotPtr
}

function getOutlines(doc, rootOutlinePtr, outlines, level) {
  ++level
  const ptrs = []
  const isValid = Module._IsValid(rootOutlinePtr)
  if (!isValid) {
    return outlines
  }

  const firstOutlinePtr = Module._GetFirst(rootOutlinePtr)
  if (!Module._IsValid(firstOutlinePtr)) {
    return outlines
  }

  let outlinePtr = createOutline(doc, ptrs, firstOutlinePtr, outlines, level)
  while (Module._IsValid(outlinePtr)) {
    outlinePtr = createOutline(doc, ptrs, outlinePtr, outlines, level)
  }
  for (let i = 0; i < ptrs.length; i++) {
    getOutlines(doc, ptrs[i], outlines[i].children, level)
  }
}

function createOutline(doc, ptrs, outlinePtr, outlines, level) {
  ptrs.push(outlinePtr)
  Module._GetTitle(outlinePtr)
  const outline = {
    title: U8StringData,
    children: [],
    level
  }

  const actionType = Module._GetOutlineActionType(doc, outlinePtr)

  getOutlineAction(doc, outlinePtr, actionType, outline)
  outlines.push(outline)

  return Module._GetNext(outlinePtr)
}

function getOutlineAction(doc, outlinePtr, actionType, outline) {
  if (actionType === 1) {
    PDFDestination = {}
    Module._GetOutlineDestination(doc, outlinePtr)
    outline.pageNumber = PDFDestination.PageIndex + 1
    outline.posX = PDFDestination.PosX < 0 ? 0 : PDFDestination.PosX
    outline.posY = PDFDestination.PosY < 0 ? 0 : PDFDestination.PosY
  } else if (actionType === 6) {
    Module._GetUri(outlinePtr)
    outline.url = U8StringData
  }
}

function getCommonAnntationAttr({
  pagePtr,
  annotPtr
}) {
  const lineWidth = Module._GetAnnotBorderWidth(annotPtr)
  const annotation = {
    lineWidth,
  }

  return annotation
}

function getMarkup({
  pagePtr,
  annotPtr
}) {
  RGBColor = {}
  const annotationAttr = {}
  const color = {}
  const hasColor = Module._GetAnnotRGBColor(annotPtr)
  if (hasColor) {
    for (key in RGBColor) {
      color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
    }
    annotationAttr.color = color
  }

  InkPointArray = []
  Module._GetQuadPoints(pagePtr, annotPtr)
  annotationAttr.quadPoints = InkPointArray

  return annotationAttr
}

function getInk({
  pagePtr,
  annotPtr
}) {
  const count = Module._GetInkPathCount(annotPtr)

  const inkPointes = []
  for (let i = 0; i < count; i++) {
    InkPointArray = []
    Module._GetInkPoints(pagePtr, annotPtr, i)
    inkPointes.push(InkPointArray)
  }

  const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

  return {
    lineWidth,
    inkPointes
  }
}

function getAnnotation({
  doc,
  pagePtr,
  annotPtr,
  typeInt,
  scale
}) {
  let annotationAttr = null
  Rect = {}
  Module._GetAnnotRect(pagePtr, annotPtr)
  const rect = {
    left: roundToDecimalPlaces(Rect.Left),
    top: roundToDecimalPlaces(Rect.Top),
    right: roundToDecimalPlaces(Rect.Right),
    bottom: roundToDecimalPlaces(Rect.Bottom)
  }

  Module._GetAnnotCreationDate(annotPtr)
  const rawCreateDate = U8StringData
  const createDate = parseAdobePDFTimestamp(rawCreateDate)
  Module._GetAnnotContent(annotPtr)
  const content = U8StringData
  Module._GetAnnotTitle(annotPtr)
  const title = U8StringData

  const transparency = roundToDecimalPlaces(Module._GetAnnotTransparency(annotPtr))

  const annotation = {
    createDate,
    content,
    rect,
    transparency
  }

  let hasColor = 0
  let isFilled = 0
  let hasBgColor = 0
  let fillTransparency = null
  switch (typeInt) {
    case AnnotationType.TEXT:
      RGBColor = {}
      Module._GetAnnotRGBColor(annotPtr)
      annotation.color = {
        R: roundToDecimalPlaces(RGBColor.R * 255),
        G: roundToDecimalPlaces(RGBColor.G * 255),
        B: roundToDecimalPlaces(RGBColor.B * 255)
      }

      return annotation
    case AnnotationType.LINK:
      const actionType = Module._GetLinkActionType(annotPtr)
      annotation.actionType = ActionTypeString[actionType]
      RGBColor = {}
      isFilled = Module._GetAnnotFilledRGBColor(annotPtr)
      if (isFilled) {
        const fillColor = {}
        for (key in RGBColor) {
          fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.fillColor = fillColor
      }
      if (actionType === 1) {
        PDFDestination = {}
        Module._GetLinkDestination(doc, annotPtr)
        annotation.destPage = PDFDestination.PageIndex + 1
        annotation.url = ''
      } else if (actionType === 6) {
        Module._GetLinkUri(annotPtr)

        annotation.url = U8StringData
        annotation.destPage = ''
      }
      return annotation
    case AnnotationType.FREETEXT:
      FontDa = {}
      Module._GetFreeTextFont(annotPtr)
      const fontName = U8StringData
      annotation.fontName = fontName
      annotation.fontSize = FontDa.fontSize

      const textAlignment = Module._GetTextAlignment(annotPtr)
      annotation.textAlignment = textAlignment

      annotation.color = {
        R: roundToDecimalPlaces(FontDa.R * 255),
        G: roundToDecimalPlaces(FontDa.G * 255),
        B: roundToDecimalPlaces(FontDa.B * 255),
        A: FontDa.A
      }

      Module._GetFreeTextContent(annotPtr)
      const freetextContent = U8StringData
      annotation.content = freetextContent

      RGBColor = {}
      hasBgColor = Module._GetFreeTextBgColor(annotPtr)
      if (hasBgColor) {
        const bgColor = {}
        for (key in RGBColor) {
          bgColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.bgColor = bgColor
      }
      return annotation
    case AnnotationType.LINE:
      LinePoint = {}
      Module._GetLinePoint(pagePtr, annotPtr)
      const linePoints = [LinePoint.StartPointX, LinePoint.StartPointY, LinePoint.EndPointX, LinePoint.EndPointY]
      annotation.linePoints = linePoints

      LineType = {}
      Module._GetLineType(annotPtr)
      annotation.tail = LineTypeString[LineType.Tail],
        annotation.head = LineTypeString[LineType.Head]

      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      annotation.lineWidth = Module._GetAnnotBorderWidth(annotPtr)
      return annotation
    case AnnotationType.SQUARE:
    case AnnotationType.CIRCLE:
      const lineWidth = Module._GetAnnotBorderWidth(annotPtr)

      annotation.lineWidth = lineWidth

      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      RGBColor = {}
      isFilled = Module._GetAnnotFilledRGBColor(annotPtr)
      if (isFilled) {
        const fillColor = {}
        for (key in RGBColor) {
          fillColor[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }
        annotation.fillColor = fillColor
      }

      fillTransparency = Module._GetAnnotFilledTransparency(annotPtr)
      annotation.fillTransparency = fillTransparency

      return annotation
    case AnnotationType.HIGHLIGHT:
    case AnnotationType.UNDERLINE:
    case AnnotationType.SQUIGGLY:
    case AnnotationType.STRIKEOUT:
      annotationAttr = getMarkup({
        pagePtr,
        annotPtr
      })
      Object.assign(annotation, annotationAttr)
      return annotation
    case AnnotationType.STAMP:
      const stampType = Module._GetStampType(annotPtr)
      annotation.stampType = StampTypeString[stampType]
      if (stampType === 0) {
        annotation.stampType = 'image'
      } else {
        TextStamp = {}
        Module._GetTextStamp(annotPtr)
        annotation.content = TextStamp.Content
        annotation.stampColor = TextStamp.StampColor
        annotation.stampShape = TextStampShapeString[TextStamp.StampType]
        if (TextStamp.Date) {
          annotation.date = TextStamp.Date
        }
      }
      return annotation
    case AnnotationType.INK:
      RGBColor = {}
      hasColor = Module._GetAnnotRGBColor(annotPtr)
      if (hasColor) {
        const color = {}
        for (key in RGBColor) {
          color[key] = roundToDecimalPlaces(RGBColor[key] * 255)
        }

        annotation.color = color
      }

      annotationAttr = getInk({
        pagePtr,
        annotPtr
      })
      Object.assign(annotation, annotationAttr)
      return annotation
  }

  return
}

function getWidgetAnnotation({
  doc,
  pagePtr,
  annotPtr,
  typeInt
}) {
  // WidgetType
  let annotationAttr = null
  Rect = {}
  Module._GetAnnotRect(pagePtr, annotPtr)
  const rect = {
    left: roundToDecimalPlaces(Rect.Left),
    top: roundToDecimalPlaces(Rect.Top),
    right: roundToDecimalPlaces(Rect.Right),
    bottom: roundToDecimalPlaces(Rect.Bottom)
  }

  Module._GetAnnotCreationDate(annotPtr)
  const rawCreateDate = U8StringData
  const createDate = parseAdobePDFTimestamp(rawCreateDate)
  Module._GetAnnotContent(annotPtr)
  const content = U8StringData

  const type = Module._GetWidgetType(annotPtr)

  const annotation = {
    createDate,
    content,
    rect,
    type: WidgetTypeString[type]
  }

  RGBColor = {}
  const hasColor = Module._GetWidgetBackgroundRGBColor(annotPtr)
  if (hasColor) {
    const { R, G, B } = RGBColor
    const hex = convertCppRGBToHex(R, G, B)
    annotation.backgroundColor = hex
  } else {
    annotation.backgroundColor = null
  }

  RGBColor = {}
  const hasBorderColor = Module._GetWidgetBorderRGBColor(annotPtr)
  if (hasBorderColor) {
    const { R, G, B } = RGBColor
    const hex = convertCppRGBToHex(R, G, B)
    annotation.borderColor = hex
  } else {
    annotation.borderColor = null
  }

  const borderStyleType = Module._GetWidgetBorderStyle(annotPtr)
  annotation.borderStyle = BorderStyleString[borderStyleType]

  Module._GetWidgetFieldName(annotPtr)
  const fieldName = U8StringData
  annotation.fieldName = fieldName

  const borderWidth = Module._GetAnnotBorderWidth(annotPtr)
  annotation.borderWidth = borderWidth

  const flag = Module._GetAnnotFlag(annotPtr)

  annotation.flag = flag
  if (flag === 0) {
    annotation.isHidden = 0
  } else {
    const invisible = !(flag & AnnotationFlags.AnnotationFlagInvisible)
    const noView = !(flag & AnnotationFlags.AnnotationFlagNoView)
    const hidden = !(flag & AnnotationFlags.AnnotationFlagHidden)
    annotation.isHidden = !(invisible && noView && hidden) ? 1 : 0
  }
  if (type === 1) {
    const checkStyle = Module._GetWidgetCheckStyle(annotPtr)
    annotation.checkStyle = checkStyle

    RGBColor = {}
    const hasColor = Module._GetWidgetCheckRGBColor(annotPtr)
    if (hasColor) {
      const checkRGBColor = {
        R: roundToDecimalPlaces(RGBColor.R * 255),
        G: roundToDecimalPlaces(RGBColor.G * 255),
        B: roundToDecimalPlaces(RGBColor.B * 255)
      }
      annotation.checkRGBColor = checkRGBColor
    }

    const isChecked = Module._GetCheckBoxIsChecked(annotPtr)
    annotation.isChecked = isChecked
  }

  if (type === 2) {
    const checkStyle = Module._GetWidgetCheckStyle(annotPtr)
    annotation.checkStyle = checkStyle
    const isChecked = Module._GetRadioButtonIsChecked(annotPtr)
    annotation.isChecked = isChecked
  }

  if (type === 3) {
    const textFieldSpe = Module._GetTextFieldSpe(annotPtr)
    if (textFieldSpe === 1) {
      return null
    }

    Module._GetTextFieldContent(annotPtr)
    const content = U8StringData
    annotation.content = content

    const isMultiLine = Module._IsMultiLine(annotPtr)
    annotation.isMultiLine = isMultiLine

    const align = Module._GetWidgetTextJustification(annotPtr)
    annotation.textAlignment = ALIGN[align]

    FontDa = {}
    Module._GetAnnotFontDa(annotPtr)
    const fontName = U8StringData
    const isBold = /bold/i.test(fontName)
    const isItalic = /(oblique|italic)/i.test(fontName)
    const textStyle = {
      fontName: fontName.split('-')[0],
      fontSize: FontDa.fontSize,
      isBold,
      isItalic
    }
    Object.assign(annotation, textStyle)

    const { R, G, B } = FontDa
    const hex = convertCppRGBToHex(R, G, B)
    annotation.color = hex
  }

  if (type === 0) {
    Module._GetButtonTitle(annotPtr)
    const title = U8StringData
    annotation.title = title

    const actionType = Module._GetBtnActionType(annotPtr)

    Module._GetBtnActionNamedAction(annotPtr)
    const actionName = U8StringData
    annotation.actionName = actionName
    if (actionType === 1) {
      const destPage = Module._GetBtnActionDestPage(doc, annotPtr)
      annotation.destPage = destPage + 1

    } else if (actionType === 6) {
      Module._GetBtnActionURI(annotPtr)
      const actionURI = U8StringData
      annotation.url = actionURI
    }
    annotation.actionType = actionType

    FontDa = {}
    Module._GetAnnotFontDa(annotPtr)
    const fontName = U8StringData
    const isBold = /bold/i.test(fontName)
    const isItalic = /(oblique|italic)/i.test(fontName)
    const textStyle = {
      fontName: fontName.split('-')[0],
      fontSize: FontDa.fontSize,
      isBold,
      isItalic
    }
    Object.assign(annotation, textStyle)

    const { R, G, B } = FontDa
    const hex = convertCppRGBToHex(R, G, B)
    annotation.color = hex
  }

  if (type === 4 || type === 5) {
    WidgetItem = {}
    Module._GetItemsSelectedItem(annotPtr)
    annotation.selectedItem = {
      values: WidgetItem.Values,
      strings: WidgetItem.Strings
    }

    const itemsCount = Module._GetItemsCount(annotPtr)
    annotation.itemsCount = itemsCount
    WidgetItemsArray = []
    Module._GetWidgetItems(annotPtr)
    annotation.items = WidgetItemsArray

    Module._GetItemsSelectedValue(annotPtr)
    const selectedValue = U8StringData
    if (selectedValue) {
      for (let i = 0; i < WidgetItemsArray.length; i++) {
        if (WidgetItemsArray[i].Value === selectedValue) {
          annotation.selected = i
          break
        }
      }
    } else {
      annotation.selected = null
    }

    FontDa = {}
    Module._GetAnnotFontDa(annotPtr)
    const fontName = U8StringData
    const isBold = /bold/i.test(fontName)
    const isItalic = /(oblique|italic)/i.test(fontName)
    const textStyle = {
      fontName: fontName.split('-')[0],
      fontSize: FontDa.fontSize,
      isBold,
      isItalic
    }
    Object.assign(annotation, textStyle)

    const { R, G, B } = FontDa
    const hex = convertCppRGBToHex(R, G, B)
    annotation.color = hex
  }

  return annotation
}

async function copyDocument(doc, password) {
  DataArray = []
  Module._SaveDocumentByStream(doc, 2);
  let pdfData = []
  for (let i = 0; i < DataArray.length; i++) {
    let subArray = new Uint8Array(DataArray[i])
    pdfData.push(subArray)
  }
  DataArray = []
  const blobData = new Blob(pdfData, { type: "application/pdf" })

  const buffer = await convertFileToBuffer(blobData)

  const newDoc = Module._InitDocument()
  const length = ComPDFKitJS.opened_files.length
  ComPDFKitJS.opened_files[length] = buffer

  const passwordPtr = stringToNewUTF8(password)
  const a = Module._LoadDocumentByStream(newDoc, length, buffer.length, passwordPtr)
  return newDoc
}

function saveDocument(doc, saveType) {
  DataArray = []
  Module._SaveDocumentByStream(doc, saveType);
  let pdfData = []
  for (let i = 0; i < DataArray.length; i++) {
    let subArray = new Uint8Array(DataArray[i])
    pdfData.push(subArray)
  }
  DataArray = []
  const blobData = new Blob(pdfData, { type: "application/pdf" })
  return blobData
}

async function saveDocumentByStream(data) {
  const { doc, saveType = 2, password, oldPassword } = data
  if (saveType === 2) {
    if (oldPassword === password || !password) {
      return saveDocument(doc, saveType)
    }
    if (password) {
      const newDoc = await copyDocument(doc, oldPassword)
      setPassword({
        doc: newDoc,
        password
      })
      const blobData = saveDocument(newDoc, saveType)
      ComPDFKitJS.opened_files.pop()
      return blobData
    } else {
      return saveDocument(doc, saveType)
    }
  } else {
    const blobData = saveDocument(doc, saveType)
    return blobData
  }
}

async function flattenPdfDownload(data) {
  const { doc, saveType = 2, password = '', oldPassword = '' } = data
  debugger
  const newDoc = await copyDocument(doc, oldPassword)

  const pagesCount = Module._GetPageCount(newDoc)
  const pagesPtr = []
  for (let i = 0; i < pagesCount; i++) {
    const pagePtr = Module._GetPage(newDoc, i)
    pagesPtr.push(pagePtr)
  }

  for (let i = 0; i < pagesPtr.length; i++) {
    Module._FlattenPage(pagesPtr[i])
  }

  if (password && password !== oldPassword) {
    setPassword({
      doc: newDoc,
      password
    })
    return saveDocument(newDoc, saveType)
  }
  const blobData = saveDocument(newDoc, saveType)
  ComPDFKitJS.opened_files.pop()
  return blobData
}

function exportXFDFByStream(doc) {
  XFDFData = {}
  Module._XFDFExportAnnotations(doc)
  return XFDFData
}

function setAnnotCreationDate(data) {
  const { annotPtr, createDate: rawCreateDate } = data

  const createDate = stringToNewUTF8(convertToPDFTimestamp(rawCreateDate))
  Module._SetAnnotCreationDate(annotPtr, createDate)
}

function setAnnotRect(data) {
  const { pagePtr, annotPtr, rect } = data
  const { left, right, bottom, top } = rect

  const result = Module._SetAnnotRect(pagePtr, annotPtr, left, right, bottom, top)
  return result
}

function setAnnotContent(data) {
  const { annotPtr, content: rawContent } = data

  const content = stringToNewUTF8(rawContent)
  const result = Module._SetAnnotContent(annotPtr, content)

  Module._GetAnnotContent(annotPtr)
  const res = U8StringData

  return result
}

function setAnnotRGBColor(data) {
  const { annotPtr, color: rawColor } = data
  const color = convertColorToCppFormat(rawColor)

  const isSetColorSuccess = Module._SetAnnotRGBColor(annotPtr, color.R, color.G, color.B)
  return !!isSetColorSuccess
}

function setAnnotTransparency(data) {
  const { annotPtr, transparency = 1 } = data

  const result = Module._SetAnnotTransparency(annotPtr, transparency)
  return result
}

function setAnnotBorderWidth(data) {
  const { annotPtr, lineWidth } = data

  Module._SetAnnotBorderWidth(annotPtr, lineWidth)
}

function SetAnnotFilledTransparency(data) {
  const { annotPtr, transparency = 1 } = data
  const result = Module._SetAnnotFilledTransparency(annotPtr, transparency)
  return result
}

function setQuadPoints(data) {
  const { pagePtr, annotPtr, quadPoints } = data
  const quadPointsArray = []
  for (let i = 0; i < quadPoints.length; i++) {
    const quadPoint = quadPoints[i]
    quadPointsArray.push(quadPoint.PointX, quadPoint.PointY)
  }

  const quadPointsArrayLength = quadPointsArray.length
  let quadPointsArrayBytes = quadPointsArrayLength * 4
  let quadPointsArrayPtr = _malloc(quadPointsArrayBytes)
  for (var i = 0; i < quadPointsArrayLength; i++) {
    Module.HEAPF32[quadPointsArrayPtr / 4 + i] = quadPointsArray[i]
  }

  const result = Module._SetQuadPoints(pagePtr, annotPtr, quadPointsArrayPtr, quadPointsArrayLength)
  _free(quadPointsArrayPtr)
  return result
}

function setLinePoint(data) {
  const { pagePtr, annotPtr, linePoints } = data

  const result = Module._SetLinePoint(pagePtr, annotPtr, linePoints[0], linePoints[1], linePoints[2], linePoints[3])
  return result
}

function setInkPointes(data) {
  const { pagePtr, annotPtr, inkPointes } = data

  const inkPointesArray = []
  const inklistArray = []
  const lineLength = inkPointes.length
  for (let i = 0; i < lineLength; i++) {
    const length = inkPointes[i].length
    for (let j = 0; j < length; j++) {
      inkPointesArray.push(inkPointes[i][j].PointX, inkPointes[i][j].PointY)
    }
    inklistArray.push(length)
  }

  const inklistArrayLength = inklistArray.length
  let inklistArrayBytes = inklistArrayLength * 4
  let inklistArrayPtr = _malloc(inklistArrayBytes)
  for (var i = 0; i < inklistArrayLength; i++) {
    Module.HEAP32[inklistArrayPtr / 4 + i] = inklistArray[i]
  }

  const inkPointesArrayLength = inkPointesArray.length
  let inkPointesArrayBytes = inkPointesArrayLength * 4
  let inkPointesArrayPtr = _malloc(inkPointesArrayBytes)
  for (var i = 0; i < inkPointesArrayLength; i++) {
    Module.HEAPF32[inkPointesArrayPtr / 4 + i] = inkPointesArray[i]
  }

  const res = Module._SetInkPath(pagePtr, annotPtr, inklistArrayPtr, lineLength, inkPointesArrayPtr)
  _free(inkPointesArrayPtr)
}

function setLinkDest(data) {
  const { doc, annotPtr, destPage, pageHeight } = data
  Module._SetLinkDestination(doc, annotPtr, destPage - 1, 0, pageHeight, 0)
}

function setLinkUri(data) {
  const { annotPtr, url: rawUrl } = data
  const url = stringToNewUTF8(rawUrl)
  const res = Module._SetLinkUri(annotPtr, url)
}

function setAnnotHidden(data) {
  const { annotPtr, isHidden } = data

  let flag = 0
  if (isHidden == 1) {
    flag = 2
  } else {
    flag = 4
  }
  const result = Module._SetAnnotFlag(annotPtr, flag)
  return result
}

function setCheckBoxIsChecked(data) {
  const { annotPtr, isChecked } = data

  const result = Module._SetCheckBoxIsChecked(annotPtr, isChecked)
}

function setRadioButtonIsChecked(data) {
  const { annotPtr, isChecked } = data

  const result = Module._SetRadioButtonIsChecked(annotPtr, isChecked)
}

function setCheckStyle(data) {
  const { annotPtr, checkStyle } = data

  const result = Module._SetWidgetCheckStyle(annotPtr, checkStyle)
}

function setWidgetBackgroundColor(data) {
  const { annotPtr, backgroundColor: rawBackgroundColor } = data
  const backgroundColor = convertColorToCppFormat(rawBackgroundColor)

  const result = Module._SetWidgetBackgroundRGBColor(annotPtr, backgroundColor.R, backgroundColor.G, backgroundColor.B)
}

function setWidgetBorder(data) {
  const { annotPtr, borderColor: rawBorderColor, borderWidth, borderStyle } = data
  const borderColor = convertColorToCppFormat(rawBorderColor)

  Module._SetWidgetBorderRGBColor(annotPtr, borderColor.R, borderColor.G, borderColor.B)
  const borderStyleEnum = BorderStyleInt[borderStyle]
  Module._SetWidgetBorderStyle(annotPtr, borderStyleEnum)
  setAnnotBorderWidth({
    annotPtr,
    lineWidth: borderWidth
  })
}

function setWidgetFieldName(data) {
  const { annotPtr, fieldName: rawFieldName } = data
  const fieldName = stringToNewUTF8(rawFieldName)

  Module._SetWidgetFieldName(annotPtr, fieldName)
}

function setAnnotFontData(data) {
  const { annotPtr, fontName: rawFontName, fontSize, color: rawColor } = data
  const fontName = stringToNewUTF8(rawFontName)
  const color = convertColorToCppFormat(rawColor)

  const result = Module._SetAnnotFontDa(annotPtr, fontName, fontSize, color.R, color.G, color.B)
}

function setAnnotSelected(data) {
  const { annotPtr, selected } = data

  const result = Module._SetItemsSelectedIndex(annotPtr, selected)
}

function setPushButtonDest(data) {
  const { doc, annotPtr, destPage, pageHeight, actionType } = data
  const actionPtr = Module._InitAction()
  Module._SetActionType(actionPtr, actionType)
  if (destPage && actionType == 1) {
    Module._SetDestination(actionPtr, doc, destPage - 1, 0, pageHeight, 0)
    Module._SetButtonAction(annotPtr, actionPtr)
    Module._ClearAction(actionPtr)
  }
}
function setPushButtonUri(data) {
  const { annotPtr, url, actionType } = data
  const actionPtr = Module._InitAction()
  Module._SetActionType(actionPtr, actionType)
  if (url && actionType == 6) {
    const UriPath = stringToNewUTF8(url)
    Module._SetUriPath(actionPtr, UriPath)
    Module._SetButtonAction(annotPtr, actionPtr)
    Module._ClearAction(actionPtr)
  }
}

function setTextFieldContent(data) {
  const { annotPtr, content: rawContent } = data
  const content = stringToNewUTF8(rawContent)
  Module._SetTextFieldContent(annotPtr, content)
}

function addWidgetItem(data) {
  const { annotPtr, item } = data
  const { index, value: rawValue } = item
  const value = stringToNewUTF8(rawValue)
  Module._AddOptionItem(annotPtr, index, value, value)
}

function removeWidgetItem(data) {
  const { annotPtr, index } = data
  Module._RemoveOptionItem(annotPtr, index)
}

function setWidgetItems(data) {
  const { annotPtr, moveOption } = data
  const { fromIndex, toIndex, item } = moveOption
  removeWidgetItem({
    annotPtr,
    index: fromIndex
  })

  addWidgetItem({
    annotPtr,
    item: {
      value: item.Value,
      index: toIndex
    }
  })
}

function deleteWidgetItem(data) {
  const { annotPtr, deleteItemIndex } = data
  const res = Module._RemoveOptionItem(annotPtr, deleteItemIndex)
}

function createMarkup(data) {
  const { pagePtr, annotPtr, color, transparency, quadPoints } = data
  setAnnotRGBColor({
    annotPtr,
    color
  })

  setAnnotTransparency({
    annotPtr,
    transparency
  })

  setQuadPoints({
    pagePtr,
    annotPtr,
    quadPoints
  })
}

function setFreeText(data) {
  const { annotPtr, fontName: rawFontName, fontSize, color, textAlignment, content: rawContent } = data

  const fontName = stringToNewUTF8(rawFontName)
  const content = stringToNewUTF8(rawContent)

  Module._SetFreeText(annotPtr, fontName, fontSize, color.R / 255, color.G / 255, color.B / 255, textAlignment, content)
}

function createShape(data) {
  const { annotPtr, lineWidth } = data
  setAnnotBorderWidth({
    annotPtr,
    lineWidth
  })
}

function createStandardStamp(data) {
  const { annotPtr, content: rawContent } = data

  const content = stringToNewUTF8(rawContent)
  Module._SetStandardStamp(annotPtr, content, 0)
}

function createTextStamp(data) {
  const { annotPtr, content: rawContent, date: rawDate, stampShape, stampColor } = data

  const content = stringToNewUTF8(rawContent)
  const date = stringToNewUTF8(rawDate)

  Module._SetTextStamp(annotPtr, content, date, stampShape, stampColor, 0)
}

function createImageStamp(data) {
  const { annotPtr, imageData, stampShape, rect, doc } = data

  setAnnotContent({
    annotPtr,
    content: 'image'
  })
  // const width = parseInt(564)
  // const height = parseInt(82)
  // let imageBufferPtr = _malloc(imageData.length)
  // Module.HEAPU8.set(imageData, imageBufferPtr)

  // const res = Module._SetImageStampWithPixel(annotPtr, imageBufferPtr, width, height, 0)

  ComPDFKitJS.opened_image = []
  ComPDFKitJS.opened_image[0] = imageData
  // const res = Module._InsertPageWithImageByStream(doc, 0, width, height, imageData, imageData.length, 0)
  const res = Module._SetImageStampByStream(annotPtr, imageData, imageData.length, 0)
  // _free()
}
